const _ = require('lodash');
const stylelint = require('stylelint');
const { report, ruleMessages, validateOptions } = stylelint.utils;

// Plugin name
const ruleName = 'sds-stylelint-plugin/namespace-pattern';

// Custom message for report
const messages = ruleMessages(ruleName, {
  expected: (prop, ns) =>
    `Expected "${prop}" to begin with "${
      _.isArray(ns) ? ns.map((n) => `--${n}-`) : `--${ns}-`
    }", have a scope of "g", "s", or "c" and be kebab-case"`,
});

module.exports = stylelint.createPlugin(ruleName, function getPlugin(actual, options) {
  return function lint(postcssRoot, postcssResult) {
    const validOptions = validateOptions(
      postcssResult,
      ruleName,
      { actual },
      {
        actual: options,
        possible: {
          namespace: [_.isString, _.isArray],
          property: [_.isBoolean],
          value: [_.isBoolean],
        },
      },
    );

    if (!validOptions) {
      // If the options are invalid, don't lint
      return;
    }
    // Set namespace based on options
    const ns = _.isArray(options.namespace)
      ? `(?<=)(${options.namespace.join('|')})(?:(?=)|$)`
      : options.namespace.toLowerCase();

    // Needed to parse out custom properties when used as fallbacks
    const customPropPattern = new RegExp('(--([a-z0-9]*)-([a-z][a-z0-9]*)(-[a-z0-9]+)*)', 'g');
    // Regex to validate --{namespace}-[g|s|c]-* and kebab-case
    const sdsPattern = new RegExp(`(--${ns}-[g|s|c]-([a-z][a-z0-9]*)(-[a-z0-9]+)*)`);

    if (options.property) {
      postcssRoot.walkDecls((decl) => {
        const prop = decl.prop;
        // This conditional deals with custom properties authored as a property
        // First check that the CSS property is a custom property
        if (prop.startsWith('--')) {
          // If property matches sdsPattern then don't lint
          if (sdsPattern.test(prop)) {
            return;
          }

          // Report custom properties that don't match valid pattern
          report({
            ruleName,
            result: postcssResult,
            message: messages.expected(prop, options.namespace), // Build the reported message
            node: decl, // Specify the reported node
          });
        }
      });
    }

    if (options.value) {
      postcssRoot.walkDecls((decl) => {
        const value = decl.value;
        if (value.startsWith('var(')) {
          // Parse out CSS properties declared as a fallback value and store in array
          const allPropsAsValues = value.match(customPropPattern);
          // Get valid result if value in array doesn't match the expected naming syntax, return boolean
          const isValid = allPropsAsValues.every((value) => sdsPattern.test(value));
          // Get custom property that failed so we have it for reporting
          const prop = allPropsAsValues.find((value) => {
            if (!sdsPattern.test(value)) {
              return value;
            }
          });

          // If property matches sdsPattern then don't lint
          if (isValid) {
            return;
          }

          // Report custom properties that don't match valid pattern
          report({
            ruleName,
            result: postcssResult,
            message: messages.expected(prop, options.namespace), // Build the reported message
            node: decl, // Specify the reported node
          });
        }
      });
    }
  };
});

module.exports.ruleName = ruleName;
module.exports.messages = messages;
